from collections import namedtuple
import io
import json

from furl import furl

from django.core.handlers.wsgi import WSGIRequest
from django.http.request import QueryDict
from django.template import Variable, VariableDoesNotExist
from django.test.client import MULTIPART_CONTENT
from django.urls import resolve
from django.urls.exceptions import Resolver404

from mayan.apps.organizations.settings import setting_organization_url_base_path
from mayan.apps.templating.classes import Template

from .literals import API_VERSION


class BatchResponse:
    def __init__(self, name, status_code, data, headers):
        self.name = name
        self.status_code = status_code
        self.data = data
        self.headers = headers


class NestableLazyIterator:
    def __init__(
        self, iterable_string, context, context_list_index, parent_iterator=None
    ):
        self.iterable_string = iterable_string
        self.context = context
        self.context_list_index = context_list_index
        self.parent_iterator = parent_iterator
        self.items = None
        self.index = 0

    def __iter__(self):
        return self

    def __next__(self):
        # Setup the initial values on the initial access.
        if not self.items:
            if self.parent_iterator:
                next(self.parent_iterator)

            self.update_iterable_object()

        if self.index == len(self.items):
            self.index = 0

            if self.parent_iterator:
                next(self.parent_iterator)
            else:
                raise StopIteration

            self.update_iterable_object()

        value = self.items[self.index]
        self.context['iterables'][self.context_list_index] = value
        self.index += 1

        return value

    def update_iterable_object(self):
        self.items = Variable(var=self.iterable_string).resolve(context=self.context)


RenderedContent = namedtuple(
    typename='RenderedContent', field_names=(
        'body', 'include', 'method', 'name', 'url'
    )
)


class BatchRequest:
    def __init__(
        self, collection, name, url, body=None, group_name=None,
        include='true', iterables=None, method='GET'
    ):
        self.collection = collection
        self.body = body or {}
        self.include = include
        self.group_name = group_name
        self.iterables = iterables
        self.method = method
        self.name = name
        self.url = url

    def execute(self):
        if self.iterables:
            # Initialize the iterables list to allow using any index.
            self.collection.context['iterables'] = [None] * len(self.iterables)

            iterator = None
            for iterable_index, iterable in enumerate(self.iterables):
                iterator = NestableLazyIterator(
                    context=self.collection.context,
                    context_list_index=iterable_index,
                    iterable_string=iterable, parent_iterator=iterator
                )

            while True:
                try:
                    next(iterator)
                except StopIteration:
                    break
                except VariableDoesNotExist as exception:
                    self.collection.responses[self.name] = {
                        'data': {'error': str(exception)},
                        'include': 'true',
                        'is_response': True
                    }
                    return
                else:
                    rendered_content = self.render_content()

                    BatchRequest(
                        collection=self.collection,
                        body=rendered_content.body,
                        group_name=self.group_name,
                        include=rendered_content.include,
                        method=rendered_content.method,
                        name=rendered_content.name,
                        url=rendered_content.url
                    ).execute()
        else:
            rendered_content = self.render_content()

            url_parts = furl(rendered_content.url)

            try:
                resolver_match = resolve(path=url_parts.pathstr)
            except Resolver404 as exception:
                self.collection.responses[rendered_content.name] = {
                    'data': {
                        'error': '"{}" not found'.format(exception.args[0]['path'])
                    },
                    'include': 'true',
                    'is_response': True,
                    'status_code': 404
                }
                return
            else:
                environ = getattr(
                    self.collection.view_request, 'environ', {}
                ).copy()

                environ['REQUEST_METHOD'] = rendered_content.method
                environ['PATH_INFO'] = self.url
                environ['QUERY_STRING'] = url_parts.querystr

                post_query_dict = QueryDict(mutable=True)
                post_query_dict.update(rendered_content.body)
                json_body = json.dumps(post_query_dict)
                request_data = json_body.encode('utf-8')
                environ['wsgi.input'] = io.BytesIO(request_data)
                environ['CONTENT_LENGTH'] = str(len(request_data))

                if rendered_content.method == 'POST':
                    environ['CONTENT_TYPE'] = MULTIPART_CONTENT
                else:
                    environ['CONTENT_TYPE'] = 'application/octet-stream'

                request = WSGIRequest(environ=environ)
                request.LANGUAGE_CODE = getattr(
                    self.collection.view_request, 'LANGUAGE_CODE', None
                )
                request.POST = post_query_dict
                request._read_started = True
                request.auth = getattr(
                    self.collection.view_request, 'auth', None
                )
                request.csrf_processing_done = True
                request.session = getattr(
                    self.collection.view_request, 'session', None
                )
                request.user = getattr(
                    self.collection.view_request, 'user', None
                )

                response = resolver_match.func(
                    request=request, **resolver_match.kwargs
                )

                result = {
                    'data': response.data,
                    'headers': {key: value for key, value in response.items()},
                    'include': rendered_content.include,
                    'is_response': True,
                    'status_code': response.status_code
                }

                self.collection.context[rendered_content.name] = result
                self.collection.responses[rendered_content.name] = result

            if self.group_name:
                self.collection.context.setdefault('groups', {})
                self.collection.context['groups'].setdefault(
                    self.group_name, []
                )
                self.collection.context['groups'][self.group_name].append(
                    result
                )

    def render_content(self):
        rendered_body = {}
        for key, value in self.body.items():
            rendered_key = Template(template_string=key).render(
                context=self.collection.context
            )
            rendered_value = Template(template_string=value).render(
                context=self.collection.context
            )
            rendered_body[rendered_key] = rendered_value

        rendered_include = Template(template_string=self.include).render(
            context=self.collection.context
        )
        rendered_method = Template(template_string=self.method).render(
            context=self.collection.context
        )
        rendered_name = Template(template_string=self.name).render(
            context=self.collection.context
        )
        rendered_url = Template(template_string=self.url).render(
            context=self.collection.context
        )

        return RenderedContent(
            body=rendered_body, include=rendered_include,
            method=rendered_method, name=rendered_name, url=rendered_url
        )


class BatchRequestCollection:
    def __init__(self, request_list=None):
        self.requests = []

        for request_index, request_dict in enumerate(request_list):
            request_dict.update(
                {'collection': self}
            )
            try:
                self.requests.append(BatchRequest(**request_dict))
            except Exception as exception:
                raise ValueError(
                    'Error instantiating request #{}; {}'.format(
                        request_index, exception
                    )
                ) from exception

    def execute(self, view_request):
        self.context = {'view_request': view_request}
        self.responses = {}
        self.view_request = view_request

        for request in self.requests:
            request.execute()

        # Convert responses in context into response class instances.
        result = []
        for key, value in self.responses.items():
            if json.loads(s=value.get('include', 'true')):
                result.append(
                    BatchResponse(
                        name=key,
                        status_code=value.get('status_code', 0),
                        data=value.get('data', {}),
                        headers=value.get('headers', {}),
                    )
                )

        return result


class Endpoint:
    def __init__(self, label, viewname=None, kwargs=None):
        self.label = label
        self.kwargs = kwargs

        if viewname:
            self.viewname = viewname
        else:
            installation_base_url = setting_organization_url_base_path.value
            if installation_base_url:
                installation_base_url = '/{}'.format(installation_base_url)
            else:
                installation_base_url = ''

            self.url = '{}/api/v{}/{}/'.format(
                installation_base_url, API_VERSION, self.label
            )

            try:
                self.viewname = resolve(path=self.url).view_name
            except Resolver404:
                self.viewname = None



import logging

from django.apps import apps
from django.db.models.signals import post_delete, post_save
from django.utils.translation import ugettext_lazy as _

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.acls.permissions import (
    permission_acl_edit, permission_acl_view
)
from mayan.apps.common.apps import MayanAppConfig
from mayan.apps.common.classes import (
    ModelCopy, ModelFieldRelated, ModelProperty, ModelQueryFields
)
from mayan.apps.common.menus import (
    menu_list_facet, menu_multi_item, menu_object, menu_related,
    menu_secondary, menu_setup
)
from mayan.apps.documents.links.document_type_links import link_document_type_list
from mayan.apps.documents.signals import signal_post_document_type_change
from mayan.apps.events.classes import EventModelRegistry, ModelEventType
from mayan.apps.navigation.classes import SourceColumn
from mayan.apps.views.html_widgets import TwoStateWidget

from .classes import DocumentMetadataHelper
from .events import (
    event_document_metadata_added, event_document_metadata_edited,
    event_document_metadata_removed, event_metadata_type_edited,
    event_metadata_type_relationship_updated
)
from .handlers import (
    handler_index_document, handler_post_document_type_metadata_type_add,
    handler_post_document_type_metadata_type_delete,
    handler_post_document_type_change_metadata
)
from .html_widgets import DocumentMetadataWidget
from .links import (
    link_metadata_add, link_metadata_edit, link_metadata_multiple_add,
    link_metadata_multiple_edit, link_metadata_multiple_remove,
    link_metadata_remove, link_metadata_view,
    link_document_type_metadata_type_relationship, link_metadata_type_create,
    link_metadata_type_delete_multiple, link_metadata_type_delete_single,
    link_metadata_type_document_type_relationship, link_metadata_type_edit,
    link_metadata_type_list
)
from .methods import method_document_get_metadata
from .permissions import (
    permission_document_metadata_add, permission_document_metadata_edit,
    permission_document_metadata_remove, permission_document_metadata_view,
    permission_metadata_type_delete, permission_metadata_type_edit,
    permission_metadata_type_view
)

logger = logging.getLogger(name=__name__)


class MetadataApp(MayanAppConfig):
    app_namespace = 'metadata'
    app_url = 'metadata'
    has_rest_api = True
    has_static_media = True
    has_tests = True
    name = 'mayan.apps.metadata'
    verbose_name = _('Metadata')

    def ready(self):
        super().ready()

        Document = apps.get_model(
            app_label='documents', model_name='Document'
        )
        DocumentFileSearchResult = apps.get_model(
            app_label='documents', model_name='DocumentFileSearchResult'
        )
        DocumentFilePageSearchResult = apps.get_model(
            app_label='documents', model_name='DocumentFilePageSearchResult'
        )
        DocumentVersionSearchResult = apps.get_model(
            app_label='documents', model_name='DocumentVersionSearchResult'
        )
        DocumentVersionPageSearchResult = apps.get_model(
            app_label='documents',
            model_name='DocumentVersionPageSearchResult'
        )

        DocumentType = apps.get_model(
            app_label='documents', model_name='DocumentType'
        )

        DocumentMetadata = self.get_model(model_name='DocumentMetadata')
        DocumentTypeMetadataType = self.get_model(
            model_name='DocumentTypeMetadataType'
        )
        MetadataType = self.get_model(model_name='MetadataType')

        Document.add_to_class(
            name='metadata_value_of', value=DocumentMetadataHelper.constructor
        )
        Document.add_to_class(
            name='get_metadata', value=method_document_get_metadata
        )

        EventModelRegistry.register(model=MetadataType)
        EventModelRegistry.register(model=DocumentTypeMetadataType)

        ModelCopy(
            model=DocumentTypeMetadataType,
        ).add_fields(
            field_names=(
                'document_type', 'metadata_type', 'required'
            )
        )

        ModelCopy(
            model=MetadataType, bind_link=True, register_permission=True
        ).add_fields(
            field_names=(
                'name', 'label', 'default', 'lookup', 'validation', 'parser',
                'document_types'
            ),
        )

        ModelProperty(
            model=Document, name='metadata_value_of.< metadata type name >',
            description=_(
                'Return the value of a specific document metadata.'
            ), label=_('Metadata value of')
        )

        ModelFieldRelated(
            model=Document, name='metadata__metadata_type__name',
            label=_('Metadata type name')
        )
        ModelFieldRelated(
            model=Document, name='metadata__value',
            label=_('Metadata type value')
        )

        ModelEventType.register(
            model=Document, event_types=(
                event_document_metadata_added,
                event_document_metadata_edited,
                event_document_metadata_removed,
            )
        )

        ModelEventType.register(
            model=MetadataType, event_types=(
                event_document_metadata_added,
                event_document_metadata_edited,
                event_document_metadata_removed,
                event_metadata_type_edited,
                event_metadata_type_relationship_updated
            )
        )

        ModelEventType.register(
            model=DocumentType, event_types=(
                event_metadata_type_relationship_updated,
            )
        )

        ModelPermission.register(
            model=Document, permissions=(
                permission_document_metadata_add,
                permission_document_metadata_edit,
                permission_document_metadata_remove,
                permission_document_metadata_view
            )
        )
        ModelPermission.register(
            model=MetadataType, permissions=(
                permission_acl_edit, permission_acl_view,
                permission_document_metadata_add,
                permission_document_metadata_edit,
                permission_document_metadata_remove,
                permission_document_metadata_view,
                permission_metadata_type_delete,
                permission_metadata_type_edit, permission_metadata_type_view
            )
        )
        ModelPermission.register_inheritance(
            model=DocumentMetadata, related='metadata_type',
        )

        model_query_fields_document = ModelQueryFields(model=Document)
        model_query_fields_document.add_prefetch_related_field(
            field_name='metadata'
        )

        # Columns

        SourceColumn(
            source=Document, label=_('Metadata'),
            widget=DocumentMetadataWidget
        )

        SourceColumn(
            attribute='document', source=DocumentFileSearchResult,
            label=_('Metadata'), widget=DocumentMetadataWidget
        )
        SourceColumn(
            attribute='document_file__document',
            source=DocumentFilePageSearchResult, label=_('Metadata'),
            widget=DocumentMetadataWidget
        )
        SourceColumn(
            attribute='document', source=DocumentVersionSearchResult,
            label=_('Metadata'), widget=DocumentMetadataWidget
        )
        SourceColumn(
            attribute='document_version__document',
            source=DocumentVersionPageSearchResult, label=_('Metadata'),
            widget=DocumentMetadataWidget
        )

        SourceColumn(
            attribute='metadata_type', is_identifier=True,
            is_sortable=True, source=DocumentMetadata
        )
        SourceColumn(
            attribute='value', include_label=True, is_sortable=True,
            source=DocumentMetadata
        )

        SourceColumn(
            attribute='is_required', include_label=True,
            source=DocumentMetadata, widget=TwoStateWidget
        )

        SourceColumn(
            attribute='label', is_identifier=True, is_sortable=True,
            source=MetadataType
        )
        SourceColumn(
            attribute='name', include_label=True, is_sortable=True,
            source=MetadataType
        )

        # Document metadata

        menu_list_facet.bind_links(
            links=(link_metadata_view,), sources=(Document,)
        )
        menu_multi_item.bind_links(
            links=(
                link_metadata_multiple_add, link_metadata_multiple_edit,
                link_metadata_multiple_remove
            ), sources=(Document,)
        )

        menu_secondary.bind_links(
            links=(
                link_metadata_add, link_metadata_edit, link_metadata_remove
            ), sources=(
                'metadata:metadata_add', 'metadata:metadata_edit',
                'metadata:metadata_remove', 'metadata:metadata_view'
            )
        )

        # Document type

        menu_list_facet.bind_links(
            links=(link_document_type_metadata_type_relationship,), sources=(
                DocumentType,
            )
        )

        menu_related.bind_links(
            links=(link_metadata_type_list,),
            sources=(
                DocumentType, 'documents:document_type_list',
                'documents:document_type_create'
            )
        )

        # Metadata type

        menu_list_facet.bind_links(
            links=(
                link_metadata_type_document_type_relationship,
            ), sources=(MetadataType,)
        )

        menu_multi_item.bind_links(
            links=(
                link_metadata_type_delete_multiple,
            ), sources=(MetadataType,)
        )

        menu_object.bind_links(
            links=(
                link_metadata_type_delete_single, link_metadata_type_edit
            ), sources=(MetadataType,)
        )

        menu_related.bind_links(
            links=(
                link_document_type_list,
            ), sources=(
                MetadataType, 'metadata:metadata_type_list',
                'metadata:metadata_type_create'
            )
        )

        menu_secondary.bind_links(
            links=(
                link_metadata_type_list,
                link_metadata_type_create
            ), sources=(
                MetadataType, 'metadata:metadata_type_list',
                'metadata:metadata_type_create'
            )
        )

        menu_setup.bind_links(links=(link_metadata_type_list,))

        # Signals

        post_delete.connect(
            dispatch_uid='metadata_handler_post_document_type_metadata_type_delete',
            receiver=handler_post_document_type_metadata_type_delete,
            sender=DocumentTypeMetadataType
        )
        post_save.connect(
            dispatch_uid='metadata_handler_post_document_type_metadata_type_add',
            receiver=handler_post_document_type_metadata_type_add,
            sender=DocumentTypeMetadataType
        )
        signal_post_document_type_change.connect(
            dispatch_uid='metadata_handler_post_document_type_change_metadata',
            receiver=handler_post_document_type_change_metadata,
            sender=Document
        )

        # Index updating

        post_delete.connect(
            dispatch_uid='metadata_handler_index_document_delete',
            receiver=handler_index_document,
            sender=DocumentMetadata
        )
        post_save.connect(
            dispatch_uid='metadata_handler_index_document_save',
            receiver=handler_index_document,
            sender=DocumentMetadata
        )



import logging
from pathlib import Path

from django.core.files import File
from django.utils.translation import ugettext_lazy as _

from mayan.apps.storage.models import SharedUploadedFile

from ..classes import SourceBackend
from ..exceptions import SourceException

from .literals import SOURCE_INTERVAL_UNCOMPRESS_CHOICES
from .mixins import (
    SourceBackendCompressedMixin, SourceBackendPeriodicMixin, SourceBaseMixin
)

__all__ = ('SourceBackendWatchFolder',)
logger = logging.getLogger(name=__name__)


class SourceBackendWatchFolder(
    SourceBackendCompressedMixin, SourceBackendPeriodicMixin,
    SourceBaseMixin, SourceBackend
):
    field_order = ('folder_path', 'include_subdirectories',)
    fields = {
        'folder_path': {
            'class': 'django.forms.CharField',
            'default': '',
            'help_text': _(
                'Server side filesystem path.'
            ),
            'kwargs': {
                'max_length': 255,
            },
            'label': _('Folder path'),
            'required': True
        },
        'include_subdirectories': {
            'class': 'django.forms.BooleanField',
            'default': '',
            'help_text': _(
                'If checked, not only will the folder path be scanned for '
                'files but also its subdirectories.'
            ),
            'label': _('Include subdirectories?'),
            'required': False
        },
    }
    label = _('Watch folder')
    uncompress_choices = SOURCE_INTERVAL_UNCOMPRESS_CHOICES

    def get_shared_uploaded_files(self):
        dry_run = self.process_kwargs.get('dry_run', False)

        path = Path(self.kwargs['folder_path'])

        # Force testing the path and raise errors for the log.
        path.lstat()
        if not path.is_dir():
            raise SourceException('Path {} is not a directory.'.format(path))

        if self.kwargs['include_subdirectories']:
            iterator = path.rglob(pattern='*')
        else:
            iterator = path.glob(pattern='*')

        for entry in iterator:
            if entry.is_file() or entry.is_symlink():
                with entry.open(mode='rb+') as file_object:
                    shared_uploaded_file = SharedUploadedFile.objects.create(
                        file=File(file=file_object), filename=entry.name
                    )
                    if not dry_run:
                        entry.unlink()

                    return (shared_uploaded_file,)



from django.template import Library
from django.urls import reverse

from ..classes import AppImageErrorImage
from ..exceptions import AppImageError

register = Library()


@register.simple_tag(takes_context=True)
def converter_get_object_image_data(
    context, obj, maximum_layer_order=None, transformation_instance_list=None,
    user=None
):
    try:
        return {
            'url': obj.get_api_image_url(
                maximum_layer_order=maximum_layer_order,
                transformation_instance_list=transformation_instance_list or (),
                user=context.get('user', user)
            )
        }
    except AppImageError as exception:
        app_image_error_image = AppImageErrorImage.get(
            name=exception.error_name
        )

        return {
            'app_image_error_image': app_image_error_image,
            'url': reverse(
                viewname='rest_api:app-image-error-image', kwargs={
                    'app_image_error_name': exception.error_name
                }
            )
        }



from shutil import copyfileobj

import magic

from mayan.apps.storage.utils import NamedTemporaryFile


def get_mimetype(file_object, mime=True, mimetype_only=False):
    """
    Determine a file's mimetype by calling the system's libmagic
    library via python-magic.
    """
    file_mimetype = None
    file_mime_encoding = None

    with NamedTemporaryFile() as temporary_file_object:
        file_object.seek(0)
        copyfileobj(fsrc=file_object, fdst=temporary_file_object)
        file_object.seek(0)
        temporary_file_object.seek(0)

        kwargs = {'mime': mime}

        if not mimetype_only:
            kwargs['mime_encoding'] = True

        mime = magic.Magic(**kwargs)

        if mimetype_only:
            file_mimetype = mime.from_file(filename=temporary_file_object.name)
        else:
            file_mimetype, file_mime_encoding = mime.from_file(
                filename=temporary_file_object.name
            ).split('; charset=')

        return file_mimetype, file_mime_encoding



import mock

import mayan

from mayan.apps.testing.tests.base import GenericViewTestCase

from ..classes import DependencyGroup
from ..literals import (
    MESSAGE_NOT_LATEST, MESSAGE_UNKNOWN_VERSION, MESSAGE_UP_TO_DATE
)
from ..permissions import permission_dependencies_view

from .literals import (
    TEST_DEPENDENCY_GROUP_NAME, TEST_DEPENDENCY_GROUP_ENTRY_NAME
)
from .mixins import (
    CheckVersionViewTestMixin, DependencyGroupEntryDetailViewTestMixin
)


class CheckVersionViewTestCase(CheckVersionViewTestMixin, GenericViewTestCase):
    @mock.patch(
        'mayan.apps.dependencies.utils.PyPIClient.get_versions', autospec=True
    )
    def test_check_version_no_permission(self, mock_package_releases):
        mock_package_releases.return_value = (mayan.__version__,)
        response = self._request_check_version_view()
        self.assertNotContains(
            response=response, text=MESSAGE_UP_TO_DATE, status_code=403
        )

    @mock.patch(
        'mayan.apps.dependencies.utils.PyPIClient.get_versions', autospec=True
    )
    def test_check_version_not_latest_version(self, mock_package_releases):
        self.grant_permission(permission=permission_dependencies_view)

        mock_package_releases.return_value = ('0.0.0',)
        response = self._request_check_version_view()
        self.assertContains(
            response=response, text=MESSAGE_NOT_LATEST[:-2], status_code=200
        )

    @mock.patch(
        'mayan.apps.dependencies.utils.PyPIClient.get_versions', autospec=True
    )
    def test_check_version_unknown_version(self, mock_package_releases):
        self.grant_permission(permission=permission_dependencies_view)

        mock_package_releases.return_value = None
        response = self._request_check_version_view()
        self.assertContains(
            response=response, text=MESSAGE_UNKNOWN_VERSION, status_code=200
        )

    @mock.patch(
        'mayan.apps.dependencies.utils.PyPIClient.get_versions', autospec=True
    )
    def test_check_version_correct_version(self, mock_package_releases):
        self.grant_permission(permission=permission_dependencies_view)

        mock_package_releases.return_value = (mayan.__version__,)
        response = self._request_check_version_view()
        self.assertContains(
            response=response, text=MESSAGE_UP_TO_DATE, status_code=200
        )


class DependencyGroupEntryDetailViewTestCase(
    DependencyGroupEntryDetailViewTestMixin, GenericViewTestCase
):
    def test_dependency_group_entry_detail_view_with_permission(self):
        test_dependency = DependencyGroup.get(
            name=TEST_DEPENDENCY_GROUP_NAME
        ).get_entry(
            entry_name=TEST_DEPENDENCY_GROUP_ENTRY_NAME
        ).get_dependencies()[0]

        self.grant_permission(permission=permission_dependencies_view)

        response = self.request_test_dependency_group_entry_detail_view()
        self.assertContains(
            response=response, text=test_dependency.name, status_code=200
        )



from django.conf.urls import url

from .views import (
    DownloadFileDeleteView, DownloadFileDownloadViewView,
    DownloadFileListView
)

urlpatterns = [
    url(
        regex=r'^downloads/(?P<download_file_id>\d+)/delete/$',
        name='download_file_delete',
        view=DownloadFileDeleteView.as_view()
    ),
    url(
        regex=r'^downloads/(?P<download_file_id>\d+)/download/$',
        name='download_file_download',
        view=DownloadFileDownloadViewView.as_view()
    ),
    url(
        regex=r'^downloads/$', name='download_file_list',
        view=DownloadFileListView.as_view()
    ),
]



import errno
from shutil import copyfile

from django.conf import settings
from django.core import management


class Command(management.BaseCommand):
    help = 'Rollback the configuration file to the last valid version.'

    def handle(self, *args, **options):
        try:
            copyfile(
                src=settings.CONFIGURATION_LAST_GOOD_FILEPATH,
                dst=settings.CONFIGURATION_FILEPATH
            )
        except IOError as exception:
            if exception.errno == errno.ENOENT:
                self.stdout.write(
                    self.style.NOTICE(
                        'There is no last valid version to restore.'
                    )
                )
            else:
                raise



from mayan.apps.documents.tests.base import GenericDocumentViewTestCase

from ..literals import WIDGET_CLASS_TEXTAREA
from ..permissions import (
    permission_workflow_instance_transition, permission_workflow_template_view
)

from .mixins.workflow_instance_mixins import WorkflowInstanceViewTestMixin
from .mixins.workflow_template_mixins import (
    WorkflowTemplateTestMixin, WorkflowTemplateViewTestMixin
)
from .mixins.workflow_template_transition_mixins import WorkflowTransitionFieldTestMixin


class WorkflowInstanceTransitionViewTestCase(
    WorkflowTemplateTestMixin, WorkflowTemplateViewTestMixin,
    WorkflowInstanceViewTestMixin, GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_workflow_template(add_test_document_type=True)
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_transition()
        self._create_test_document_stub()
        self.test_workflow_instance = self.test_document.workflows.first()

    def test_document_workflow_instance_list_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_workflow_instance_list_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_workflow_instance_list_view_with_document_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_workflow_template_view
        )

        self._clear_events()

        response = self._request_test_document_workflow_instance_list_view()
        self.assertNotContains(
            response=response, text=self.test_workflow_template.label,
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_workflow_instance_list_view_with_workflow_access(self):
        self.grant_access(
            obj=self.test_workflow_template,
            permission=permission_workflow_template_view
        )

        self._clear_events()

        response = self._request_test_document_workflow_instance_list_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_workflow_instance_list_view_with_full_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_workflow_template_view
        )
        self.grant_access(
            obj=self.test_workflow_template,
            permission=permission_workflow_template_view
        )

        self._clear_events()

        response = self._request_test_document_workflow_instance_list_view()
        self.assertContains(
            response=response, text=self.test_workflow_template.label,
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_workflow_instance_list_view_with_full_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_workflow_template_view
        )
        self.grant_access(
            obj=self.test_workflow_template,
            permission=permission_workflow_template_view
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_document_workflow_instance_list_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_detail_view_no_permission(self):
        self._clear_events()

        response = self._request_test_workflow_instance_detail_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_detail_view_with_document_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_workflow_template_view
        )

        self._clear_events()

        response = self._request_test_workflow_instance_detail_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_detail_view_with_workflow_access(self):
        self.grant_access(
            obj=self.test_workflow_template,
            permission=permission_workflow_template_view
        )

        self._clear_events()

        response = self._request_test_workflow_instance_detail_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_detail_view_with_full_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_workflow_template_view
        )
        self.grant_access(
            obj=self.test_workflow_template,
            permission=permission_workflow_template_view
        )

        self._clear_events()

        response = self._request_test_workflow_instance_detail_view()
        self.assertContains(
            response=response, text=self.test_workflow_template.label,
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_workflow_instance_detail_view_with_full_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_workflow_template_view
        )
        self.grant_access(
            obj=self.test_workflow_template,
            permission=permission_workflow_template_view
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_workflow_instance_detail_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_transition_selection_get_view_no_permission(self):
        self._clear_events()

        response = self._request_test_workflow_instance_transition_selection_get_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self.test_workflow_instance.get_current_state(),
            self.test_workflow_template_states[0]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_transition_selection_get_view_with_document_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_workflow_instance_transition
        )

        self._clear_events()

        response = self._request_test_workflow_instance_transition_selection_get_view()
        self.assertContains(
            response=response, text=str(self.test_document), status_code=200
        )
        self.assertNotContains(
            response=response,
            text=str(self.test_workflow_template_transitions[0]),
            status_code=200
        )

        self.assertEqual(
            self.test_workflow_instance.get_current_state(),
            self.test_workflow_template_states[0]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_transition_selection_get_view_with_transition_access(self):
        self.grant_access(
            obj=self.test_workflow_template_transition,
            permission=permission_workflow_instance_transition
        )

        self._clear_events()

        response = self._request_test_workflow_instance_transition_selection_get_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self.test_workflow_instance.get_current_state(),
            self.test_workflow_template_states[0]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_transition_selection_get_view_with_document_and_transition_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_workflow_instance_transition
        )
        self.grant_access(
            obj=self.test_workflow_template_transition,
            permission=permission_workflow_instance_transition
        )

        self._clear_events()

        response = self._request_test_workflow_instance_transition_selection_get_view()
        self.assertContains(
            response=response, text=str(self.test_document), status_code=200
        )
        self.assertContains(
            response=response,
            text=str(self.test_workflow_template_transitions[0]),
            status_code=200
        )

        self.assertEqual(
            self.test_workflow_instance.get_current_state(),
            self.test_workflow_template_states[0]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_transition_selection_get_view_with_workflow_access(self):
        self.grant_access(
            obj=self.test_workflow_template,
            permission=permission_workflow_instance_transition
        )

        self._clear_events()

        response = self._request_test_workflow_instance_transition_selection_get_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self.test_workflow_instance.get_current_state(),
            self.test_workflow_template_states[0]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_transition_selection_get_view_with_document_and_workflow_template_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_workflow_instance_transition
        )
        self.grant_access(
            obj=self.test_workflow_template,
            permission=permission_workflow_instance_transition
        )

        self._clear_events()

        response = self._request_test_workflow_instance_transition_selection_get_view()
        self.assertContains(
            response=response, text=str(self.test_document), status_code=200
        )
        self.assertContains(
            response=response,
            text=str(self.test_workflow_template_transitions[0]),
            status_code=200
        )

        self.assertEqual(
            self.test_workflow_instance.get_current_state(),
            self.test_workflow_template_states[0]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_workflow_instance_transition_selection_get_view_with_document_and_workflow_template_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_workflow_instance_transition
        )
        self.grant_access(
            obj=self.test_workflow_template,
            permission=permission_workflow_instance_transition
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_workflow_instance_transition_selection_get_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self.test_workflow_instance.get_current_state(),
            self.test_workflow_template_states[0]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_transition_selection_post_view_no_permission(self):
        self._clear_events()

        response = self._request_test_workflow_instance_transition_selection_post_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self.test_workflow_instance.get_current_state(),
            self.test_workflow_template_states[0]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_transition_selection_post_view_with_document_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_workflow_instance_transition
        )

        self._clear_events()

        response = self._request_test_workflow_instance_transition_selection_post_view()
        self.assertEqual(response.status_code, 200)

        self.assertEqual(
            self.test_workflow_instance.get_current_state(),
            self.test_workflow_template_states[0]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_transition_selection_post_view_with_document_and_transition_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_workflow_instance_transition
        )
        self.grant_access(
            obj=self.test_workflow_template_transition,
            permission=permission_workflow_instance_transition
        )

        self._clear_events()

        response = self._request_test_workflow_instance_transition_selection_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self.test_workflow_instance.get_current_state(),
            self.test_workflow_template_states[0]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_transition_selection_post_view_with_workflow_access(self):
        self.grant_access(
            obj=self.test_workflow_template,
            permission=permission_workflow_instance_transition
        )

        self._clear_events()

        response = self._request_test_workflow_instance_transition_selection_post_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self.test_workflow_instance.get_current_state(),
            self.test_workflow_template_states[0]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_transition_selection_post_view_with_document_and_workflow_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_workflow_instance_transition
        )
        self.grant_access(
            obj=self.test_workflow_template,
            permission=permission_workflow_instance_transition
        )

        self._clear_events()

        response = self._request_test_workflow_instance_transition_selection_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self.test_workflow_instance.get_current_state(),
            self.test_workflow_template_states[0]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_workflow_instance_transition_selection_post_view_with_document_and_workflow_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_workflow_instance_transition
        )
        self.grant_access(
            obj=self.test_workflow_template,
            permission=permission_workflow_instance_transition
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_workflow_instance_transition_selection_post_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self.test_workflow_instance.get_current_state(),
            self.test_workflow_template_states[0]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_transition_execute_view_no_permission(self):
        self._clear_events()

        response = self._request_test_workflow_instance_transition_execute_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self.test_workflow_instance.get_current_state(),
            self.test_workflow_template_states[0]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_transition_execute_view_with_document_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_workflow_instance_transition
        )

        self._clear_events()

        response = self._request_test_workflow_instance_transition_execute_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self.test_workflow_instance.get_current_state(),
            self.test_workflow_template_states[0]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_transition_execute_view_with_transition_access(self):
        self.grant_access(
            obj=self.test_workflow_template_transition,
            permission=permission_workflow_instance_transition
        )

        self._clear_events()

        response = self._request_test_workflow_instance_transition_execute_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self.test_workflow_instance.get_current_state(),
            self.test_workflow_template_states[0]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_transition_execute_view_with_document_and_transition_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_workflow_instance_transition
        )
        self.grant_access(
            obj=self.test_workflow_template_transition,
            permission=permission_workflow_instance_transition
        )

        self._clear_events()

        response = self._request_test_workflow_instance_transition_execute_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self.test_workflow_instance.get_current_state(),
            self.test_workflow_template_states[1]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_transition_execute_view_with_workflow_access(self):
        self.grant_access(
            obj=self.test_workflow_template,
            permission=permission_workflow_instance_transition
        )

        self._clear_events()

        response = self._request_test_workflow_instance_transition_execute_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self.test_workflow_instance.get_current_state(),
            self.test_workflow_template_states[0]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_transition_execute_view_with_document_and_workflow_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_workflow_instance_transition
        )
        self.grant_access(
            obj=self.test_workflow_template,
            permission=permission_workflow_instance_transition
        )

        self._clear_events()

        response = self._request_test_workflow_instance_transition_execute_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self.test_workflow_instance.get_current_state(),
            self.test_workflow_template_states[1]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_workflow_instance_transition_execute_view_with_document_and_workflow_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_workflow_instance_transition
        )
        self.grant_access(
            obj=self.test_workflow_template,
            permission=permission_workflow_instance_transition
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_workflow_instance_transition_execute_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self.test_workflow_instance.get_current_state(),
            self.test_workflow_template_states[0]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class WorkflowInstanceTransitionFieldViewTestCase(
    WorkflowTemplateTestMixin, WorkflowTransitionFieldTestMixin,
    WorkflowInstanceViewTestMixin, GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_workflow_template(add_test_document_type=True)
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_transition()
        self._create_test_workflow_template_transition_field(
            extra_data={
                'widget': WIDGET_CLASS_TEXTAREA
            }
        )
        self._create_test_document_stub()

        self.test_workflow_instance = self.test_document.workflows.first()

    def test_workflow_instance_transition_text_area_widget_execute_view_with_document_and_transition_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_workflow_instance_transition
        )
        self.grant_access(
            obj=self.test_workflow_template_transition,
            permission=permission_workflow_instance_transition
        )

        self._clear_events()

        response = self._request_test_workflow_instance_transition_execute_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self.test_workflow_instance.get_current_state(),
            self.test_workflow_template_states[1]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_workflow_instance_transition_text_area_widget_execute_view_with_document_and_transition_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_workflow_instance_transition
        )
        self.grant_access(
            obj=self.test_workflow_template_transition,
            permission=permission_workflow_instance_transition
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_workflow_instance_transition_execute_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self.test_workflow_instance.get_current_state(),
            self.test_workflow_template_states[0]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from mayan.apps.documents.tests.literals import TEST_SMALL_DOCUMENT_PATH

from django.db.models import Q

from ..models import Tag

from .literals import (
    TEST_TAG_COLOR, TEST_TAG_COLOR_EDITED, TEST_TAG_LABEL,
    TEST_TAG_LABEL_EDITED
)


class DocumentTagViewTestMixin:
    def _request_test_document_tag_attach_view(self):
        return self.post(
            viewname='tags:tag_attach', kwargs={
                'document_id': self.test_document.pk
            }, data={
                'tags': self.test_tag.pk,
                'user': self._test_case_user.pk
            }
        )

    def _request_test_document_multiple_tag_multiple_attach_view(self):
        return self.post(
            viewname='tags:multiple_documents_tag_attach', data={
                'id_list': self.test_document.pk, 'tags': self.test_tag.pk,
                'user': self._test_case_user.pk
            }
        )

    def _request_test_document_tag_multiple_remove_view(self):
        return self.post(
            viewname='tags:single_document_multiple_tag_remove', kwargs={
                'document_id': self.test_document.pk
            }, data={
                'tags': self.test_tag.pk,
            }
        )

    def _request_test_document_multiple_tag_remove_view(self):
        return self.post(
            viewname='tags:multiple_documents_selection_tag_remove', data={
                'id_list': self.test_document.pk,
                'tags': self.test_tag.pk,
            }
        )

    def _request_test_document_tag_list_view(self):
        return self.get(
            viewname='tags:document_tag_list', kwargs={
                'document_id': self.test_document.pk
            }
        )

    def _request_test_tag_document_list_view(self):
        return self.get(
            viewname='tags:tag_document_list', kwargs={
                'tag_id': self.test_tag.pk
            }
        )


class TagAPIViewTestMixin:
    def _request_test_document_tag_attach_api_view(self):
        return self.post(
            viewname='rest_api:document-tag-attach', kwargs={
                'document_id': self.test_document.pk
            }, data={'tag': self.test_tag.pk}
        )

    def _request_test_document_tag_list_api_view(self):
        return self.get(
            viewname='rest_api:document-tag-list', kwargs={
                'document_id': self.test_document.pk
            }
        )

    def _request_test_document_tag_remove_api_view(self):
        return self.post(
            viewname='rest_api:document-tag-remove', kwargs={
                'document_id': self.test_document.pk
            }, data={'tag': self.test_tag.pk}
        )

    def _request_test_tag_create_api_view(self):
        pk_list = list(Tag.objects.values('pk'))

        response = self.post(
            viewname='rest_api:tag-list', data={
                'label': TEST_TAG_LABEL, 'color': TEST_TAG_COLOR
            }
        )

        try:
            self.test_tag = Tag.objects.get(
                ~Q(pk__in=pk_list)
            )
        except Tag.DoesNotExist:
            self.test_tag = None

        return response

    def _request_test_tag_delete_api_view(self):
        return self.delete(
            viewname='rest_api:tag-detail',
            kwargs={'tag_id': self.test_tag.pk}
        )

    def _request_test_tag_detail_api_view(self):
        return self.get(
            viewname='rest_api:tag-detail',
            kwargs={'tag_id': self.test_tag.pk}
        )

    def _request_test_tag_document_list_api_view(self):
        return self.get(
            viewname='rest_api:tag-document-list', kwargs={
                'tag_id': self.test_tag.pk
            }
        )

    def _request_test_tag_edit_api_view(self, extra_data=None, verb='patch'):
        data = {
            'label': TEST_TAG_LABEL_EDITED,
            'color': TEST_TAG_COLOR_EDITED
        }

        if extra_data:
            data.update(extra_data)

        return getattr(self, verb)(
            viewname='rest_api:tag-detail',
            kwargs={'tag_id': self.test_tag.pk},
            data=data
        )

    def _request_test_tag_list_api_view(self):
        return self.get(viewname='rest_api:tag-list')


class TagTestMixin:
    def setUp(self):
        super().setUp()
        self.test_tags = []

    def _create_test_tag(self, add_test_document=False):
        total_test_labels = len(self.test_tags)
        label = '{}_{}'.format(TEST_TAG_LABEL, total_test_labels)

        self.test_tag = Tag.objects.create(
            color=TEST_TAG_COLOR, label=label
        )

        self.test_tags.append(self.test_tag)

        if add_test_document:
            self.test_tag.documents.add(self.test_document)


class TagViewTestMixin:
    def _request_test_tag_create_view(self):
        pk_list = list(Tag.objects.values('pk'))

        response = self.post(
            viewname='tags:tag_create', data={
                'label': TEST_TAG_LABEL,
                'color': TEST_TAG_COLOR
            }
        )

        try:
            self.test_tag = Tag.objects.get(
                ~Q(pk__in=pk_list)
            )
        except Tag.DoesNotExist:
            self.test_tag = None

        return response

    def _request_test_tag_delete_view(self):
        return self.post(
            viewname='tags:tag_delete_single', kwargs={
                'tag_id': self.test_tag.pk
            }
        )

    def _request_test_tag_delete_multiple_view(self):
        return self.post(
            viewname='tags:tag_delete_multiple', data={
                'id_list': self.test_tag.pk
            }
        )

    def _request_test_tag_edit_view(self):
        return self.post(
            viewname='tags:tag_edit', kwargs={
                'tag_id': self.test_tag.pk
            }, data={
                'label': TEST_TAG_LABEL_EDITED, 'color': TEST_TAG_COLOR_EDITED
            }
        )

    def _request_test_tag_list_view(self):
        return self.get(viewname='tags:tag_list')


class TaggedDocumentUploadWizardStepViewTestMixin:
    def _request_upload_interactive_document_create_view(self):
        with open(file=TEST_SMALL_DOCUMENT_PATH, mode='rb') as file_object:
            return self.post(
                viewname='sources:document_upload_interactive', kwargs={
                    'source_id': self.test_source.pk
                }, data={
                    'document_type_id': self.test_document_type.pk,
                    'source-file': file_object,
                    'tags': Tag.objects.values_list('pk', flat=True)
                }
            )



import logging

from django.apps import apps

from mayan.celery import app

logger = logging.getLogger(name=__name__)


@app.task(bind=True, ignore_result=True)
def task_unverify_key_signatures(self, key_id):
    DetachedSignature = apps.get_model(
        app_label='document_signatures', model_name='DetachedSignature'
    )

    EmbeddedSignature = apps.get_model(
        app_label='document_signatures', model_name='EmbeddedSignature'
    )

    for signature in DetachedSignature.objects.filter(key_id__endswith=key_id).filter(signature_id__isnull=False):
        signature.save()

    for signature in EmbeddedSignature.objects.filter(key_id__endswith=key_id).filter(signature_id__isnull=False):
        signature.save()


@app.task(bind=True, ignore_result=True)
def task_verify_key_signatures(self, key_pk):
    Key = apps.get_model(
        app_label='django_gpg', model_name='Key'
    )

    DetachedSignature = apps.get_model(
        app_label='document_signatures', model_name='DetachedSignature'
    )

    EmbeddedSignature = apps.get_model(
        app_label='document_signatures', model_name='EmbeddedSignature'
    )

    key = Key.objects.get(pk=key_pk)

    for signature in DetachedSignature.objects.filter(key_id__endswith=key.key_id).filter(signature_id__isnull=True):
        signature.save()

    for signature in EmbeddedSignature.objects.filter(key_id__endswith=key.key_id).filter(signature_id__isnull=True):
        signature.save()


@app.task(bind=True, ignore_result=True)
def task_verify_missing_embedded_signature(self):
    EmbeddedSignature = apps.get_model(
        app_label='document_signatures', model_name='EmbeddedSignature'
    )

    for document_file in EmbeddedSignature.objects.unsigned_document_files():
        task_verify_document_file.apply_async(
            kwargs={
                'document_file_pk': document_file.pk
            }
        )


@app.task(bind=True, ignore_result=True)
def task_verify_document_file(self, document_file_pk):
    DocumentFile = apps.get_model(
        app_label='documents', model_name='DocumentFile'
    )

    EmbeddedSignature = apps.get_model(
        app_label='document_signatures', model_name='EmbeddedSignature'
    )

    document_file = DocumentFile.objects.get(pk=document_file_pk)
    try:
        EmbeddedSignature.objects.create(document_file=document_file)
    except IOError as exception:
        error_message = 'File missing for document file ID {}; {}'.format(
            document_file_pk, exception
        )
        logger.error(error_message)
        raise IOError(error_message)


@app.task(ignore_result=True)
def task_refresh_signature_information():
    DetachedSignature = apps.get_model(
        app_label='document_signatures', model_name='DetachedSignature'
    )
    EmbeddedSignature = apps.get_model(
        app_label='document_signatures', model_name='EmbeddedSignature'
    )

    for signanture in DetachedSignature.objects.all():
        try:
            signanture.save()
        except Exception as exception:
            logger.error(
                'Error refreshing detached signature {} for document file ID {}; {}'.format(
                    signanture, signanture.document_file_id, exception,
                ), exc_info=True
            )
            raise

    for signanture in EmbeddedSignature.objects.all():
        try:
            signanture.save()
        except Exception as exception:
            logger.error(
                'Error refreshing embedded signature {} for document file ID {}; {}'.format(
                    signanture, signanture.document_file_id, exception,
                ), exc_info=True
            )
            raise



from django.utils.translation import ugettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(label=_('Appearance'), name='appearance')

event_theme_created = namespace.add_event_type(
    label=_('Theme created'), name='theme_created'
)
event_theme_edited = namespace.add_event_type(
    label=_('Theme edited'), name='theme_edited'
)



from mayan.apps.testing.tests.base import GenericViewTestCase
from mayan.apps.user_management.events import event_user_edited

from ..events import event_user_logged_in, event_user_logged_out


class UserEventsTestCase(GenericViewTestCase):
    auto_login_user = False
    create_test_case_user = False

    def test_user_logged_in_event_from_view(self):
        self._create_test_user()

        last_login = self.test_user.last_login

        self._clear_events()

        result = self.login(
            username=self.test_user.username,
            password=self.test_user.cleartext_password
        )
        self.assertTrue(result)

        self.assertEqual(self.test_user.last_login, last_login)
        self.assertEqual(self.test_user.last_login, None)

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self.test_user)
        self.assertEqual(events[0].target, self.test_user)
        self.assertEqual(events[0].verb, event_user_edited.id)

        self.assertEqual(events[1].action_object, None)
        self.assertEqual(events[1].actor, self.test_user)
        self.assertEqual(events[1].target, self.test_user)
        self.assertEqual(events[1].verb, event_user_logged_in.id)

    def test_user_logged_out_event_from_view(self):
        self._create_test_user()

        result = self.login(
            username=self.test_user.username,
            password=self.test_user.cleartext_password
        )
        self.assertTrue(result)

        self._clear_events()

        self.logout()

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self.test_user)
        self.assertEqual(events[0].target, self.test_user)
        self.assertEqual(events[0].verb, event_user_logged_out.id)



from rest_framework import serializers

from .models import DocumentFilePageContent, DocumentTypeSettings


class DocumentFilePageContentSerializer(serializers.ModelSerializer):
    class Meta:
        fields = ('content',)
        model = DocumentFilePageContent


class DocumentTypeParsingSettingsSerializer(serializers.ModelSerializer):
    class Meta:
        fields = ('auto_parsing',)
        model = DocumentTypeSettings



from .tasks import task_deindex_instance, task_index_instance


def handler_factory_deindex_instance(search_model):

    def handler_deindex_instance(sender, **kwargs):
        instance = kwargs['instance']

        task_deindex_instance.apply_async(
            kwargs={
                'app_label': instance._meta.app_label,
                'model_name': instance._meta.model_name,
                'object_id': instance.pk
            }
        )

    return handler_deindex_instance


def handler_index_instance(sender, **kwargs):
    instance = kwargs['instance']

    task_index_instance.apply_async(
        kwargs={
            'app_label': instance._meta.app_label,
            'model_name': instance._meta.model_name,
            'object_id': instance.pk
        }
    )


def handler_factory_index_instance_m2m(model):
    def handler_index_instance_m2m(sender, **kwargs):
        instance = kwargs['instance']
        action = kwargs.get('action')

        if action in ('post_add', 'post_remove'):
            if model == instance._meta.model or model == kwargs['model']._meta.model:
                task_index_instance.apply_async(
                    kwargs={
                        'app_label': instance._meta.app_label,
                        'model_name': instance._meta.model_name,
                        'object_id': instance.pk
                    }
                )

                for pk in kwargs['pk_set']:
                    task_index_instance.apply_async(
                        kwargs={
                            'app_label': kwargs['model']._meta.app_label,
                            'model_name': kwargs['model']._meta.model_name,
                            'object_id': pk
                        }
                    )

    return handler_index_instance_m2m



from django.utils.translation import ugettext_lazy as _

from mayan.apps.smart_settings.classes import SettingNamespace

from .literals import (
    DEFAULT_STORAGE_DOWNLOAD_FILE_STORAGE,
    DEFAULT_STORAGE_DOWNLOAD_FILE_STORAGE_ARGUMENTS,
    DEFAULT_STORAGE_SHARED_STORAGE, DEFAULT_STORAGE_SHARED_STORAGE_ARGUMENTS,
    DEFAULT_STORAGE_TEMPORARY_DIRECTORY
)

namespace = SettingNamespace(label=_('Storage'), name='storage')

setting_download_file_storage = namespace.add_setting(
    default=DEFAULT_STORAGE_DOWNLOAD_FILE_STORAGE,
    global_name='STORAGE_DOWNLOAD_FILE_STORAGE', help_text=_(
        'A storage backend that all workers can use to generate and hold '
        'files for download.'
    )
)
setting_download_file_storage_arguments = namespace.add_setting(
    default=DEFAULT_STORAGE_DOWNLOAD_FILE_STORAGE_ARGUMENTS,
    global_name='STORAGE_DOWNLOAD_FILE_STORAGE_ARGUMENTS',
)
setting_shared_storage = namespace.add_setting(
    default=DEFAULT_STORAGE_SHARED_STORAGE,
    global_name='STORAGE_SHARED_STORAGE', help_text=_(
        'A storage backend that all workers can use to share files.'
    )
)
setting_shared_storage_arguments = namespace.add_setting(
    default=DEFAULT_STORAGE_SHARED_STORAGE_ARGUMENTS,
    global_name='STORAGE_SHARED_STORAGE_ARGUMENTS'
)
setting_temporary_directory = namespace.add_setting(
    default=DEFAULT_STORAGE_TEMPORARY_DIRECTORY,
    global_name='STORAGE_TEMPORARY_DIRECTORY', help_text=_(
        'Temporary directory used site wide to store thumbnails, previews '
        'and temporary files.'
    )
)
